#define _WIN32_WINNT 0x0500

// C4800: 'int' : forcing value to bool 'true' or 'false' (performance warning)
#pragma warning(disable : 4800)
// Win32 templates generate hundreds of "overly-long identifier" warnings
#pragma warning(disable : 4786)
#pragma warning(disable : 4503)

#include <windows.h>
#include <commctrl.h>
#include <shlwapi.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <string>
#include <vector>
#include <set>
#include <map>

#include "resource.h"

//#define LWA_COLORKEY            0x00000001
//#define LWA_ALPHA               0x00000002
//#define WS_EX_LAYERED           0x00080000
//extern "C" BOOL SetLayeredWindowAttributes(HWND hwnd, COLORREF crKey, BYTE bAlpha, DWORD dwFlags);

using namespace std;

struct window {
    HWND hwnd;
    bool layered;
    BYTE alpha;
    bool mark;
    bool restored;
    string title;
    string classname;

    window(HWND w, const char *tit, const char *clsnm);
    window(const char *tit, const char *clsnm, BYTE alpha);

    void set_hwnd(HWND w);
};

HKEY regkey;

window *g_current = NULL;
window *g_tray = NULL;

set<window *> g_all_windows;
map<HWND, window *> g_windows_by_hwnd;
map<string, window *> g_windows_by_id;

void
load(void)
{
    char buf[1000];
    for (int num = 0; ; num++) {
	DWORD sz = sizeof(buf)-1;
	if (RegEnumKeyEx(regkey, num, buf, &sz,
			 0, NULL, 0, NULL) != ERROR_SUCCESS)
	    break;
	buf[sz] = 0;

	HKEY key;

	if (RegOpenKeyEx(regkey,
			 buf, 0,
			 KEY_ALL_ACCESS, &key) != ERROR_SUCCESS) {
	    MessageBox(NULL, NULL, "warning: Could not open registry key to restore on load ", MB_ICONWARNING | MB_OK);
	    continue;
	}

	buf[0] = 0;
	sz = sizeof(buf);
	DWORD ty;

	if (RegQueryValueEx(key, "title", NULL, &ty, (LPBYTE)buf, &sz) !=
	    ERROR_SUCCESS) {
	    MessageBox(NULL, NULL, "warning: Could not read title value from registry key to restore on load ", MB_ICONWARNING | MB_OK);
	    continue;
	}
	string title(buf, sz);

	if (RegQueryValueEx(key, "class", NULL, &ty, (LPBYTE)buf, &sz) !=
	    ERROR_SUCCESS) {
	    MessageBox(NULL, NULL, "warning: Could not read class value from registry key to restore on load ", MB_ICONWARNING | MB_OK);
	    continue;
	}
	string classname(buf, sz);

	DWORD alpha;
	sz = sizeof(alpha);
	if (RegQueryValueEx(key, "alpha", NULL, &ty, (LPBYTE)&alpha, &sz) !=
	    ERROR_SUCCESS) {
	    MessageBox(NULL, "Could not read alpha value from registry key to restore on load ", "Warning", MB_ICONWARNING | MB_OK);
	    continue;
	}

	window *win = new window(title.c_str(), classname.c_str(), alpha);

	RegCloseKey(key);
    }
}

void
save(void)
{
    int num;
    while (1) {
	char buf[1000];
	DWORD sz = sizeof(buf)-1;
	if (RegEnumKeyEx(regkey, 0, buf, &sz,
			 0, NULL, 0, NULL) != ERROR_SUCCESS)
	    break;
	buf[sz] = 0;
	if (SHDeleteKey(regkey, buf) != ERROR_SUCCESS) {
	    MessageBox(NULL, "Could not delete key", "Warning", MB_ICONWARNING | MB_OK);
	    continue;
	}
    }

    char buf[10];
    num = 1;

    set<window *>::iterator it = g_all_windows.begin();
    while (it != g_all_windows.end()) {
	window *win = *it++;
	if (win->restored) {
	    HKEY key;
	    sprintf(buf, "%d", num++);
	    if (RegCreateKeyEx(regkey,
			       buf, NULL, NULL, 0,
			       KEY_ALL_ACCESS, NULL, &key, NULL) != ERROR_SUCCESS) {
		MessageBox(NULL, NULL, "warning: Could not create registry key to restore on save ", MB_ICONWARNING | MB_OK);
		continue;
	    }
	    RegSetValueEx(key, "title", NULL, REG_SZ,
			  (LPBYTE)win->title.c_str(),
			  win->title.size());
	    RegSetValueEx(key, "class", NULL, REG_SZ,
			  (LPBYTE)win->classname.c_str(),
			  win->classname.size());
	    sprintf(buf, "%d", win->alpha);
	    DWORD d = win->alpha;
	    RegSetValueEx(key, "alpha", NULL, REG_DWORD,
			  (LPBYTE)&d, sizeof(d));
	    RegCloseKey(key);
	}
    }
}

void
set_layered(HWND hwnd, bool on)
{
    LONG l = GetWindowLong(hwnd, GWL_EXSTYLE);
    if (on) l |= WS_EX_LAYERED;
    else l &= ~(WS_EX_LAYERED);
    SetWindowLong(hwnd, GWL_EXSTYLE, l);
}

bool
is_layered(HWND hwnd, BYTE *ret_alpha)
{
    LONG l = GetWindowLong(hwnd, GWL_EXSTYLE);
    DWORD flags = LWA_ALPHA;
    if (ret_alpha) *ret_alpha = 255;
    if (l & WS_EX_LAYERED) {
	if (ret_alpha) {
	    BYTE tmp;
	    HMODULE lib = LoadLibrary("user32.dll");
	    if (!lib) {
		MessageBox(NULL, NULL, "could not load lib",
			   MB_ICONWARNING | MB_OK);
	    } else {
		BOOL
		    (*GetLayeredWindowAttributes)(HWND hwnd,
						  COLORREF *pcrKey,
						  BYTE *pbAlpha,
						  DWORD *pdwFlags);

		GetLayeredWindowAttributes =
		    (BOOL
		     (*)(HWND hwnd, COLORREF *, BYTE *, DWORD *))
		    GetProcAddress(lib, "GetLayeredWindowAttributes");
		if (!GetLayeredWindowAttributes) {
		} else {
		    GetLayeredWindowAttributes(hwnd, NULL, &tmp, &flags);
		    if (flags & LWA_ALPHA) *ret_alpha = tmp;
		}
	    }
	}
	return 1;
    }
    return 0;
}

bool
is_valid_window_to_show(HWND hwnd,
			char *buf1, size_t sz1,
			char *buf2, size_t sz2)
{
    if (!IsWindowVisible(hwnd)) return 0;

    GetWindowText(hwnd, buf1, sz1);
    GetClassName(hwnd, buf2, sz2);

    if (!strcmp(buf2, "Progman")) return 0;
    if (!strcmp(buf2, "Shell_TrayWnd")) {
	strncpy(buf1, "System Tray / Task Bar", sz1);
	return 1;
    }
    if (!strcmp(buf2, "ToolbarWindow32")) return 0;
    if (buf2[0] == '#') return 0;
    return 1;
}

int
CALLBACK
listcmp(LPARAM a, LPARAM b, LPARAM lparam)
{
    window *p = (window*)a;
    window *q = (window*)b;

    int r = stricmp(p->classname.c_str(), q->classname.c_str());
    if (r == 0)
	r = stricmp(p->title.c_str(), q->title.c_str());
    return r;
}

void
update_window_item(HWND listbox, window *win)
{
    LVFINDINFO lvfindinfo;

    lvfindinfo.flags = LVFI_PARAM;
    lvfindinfo.lParam = (LPARAM)win;

    int i = ListView_FindItem(listbox, -1, &lvfindinfo);

    if (i >= 0) {
	char buf[100];

	LVITEM lvi;
	ZeroMemory(&lvi, sizeof(lvi));
	lvi.mask = LVIF_TEXT;
	lvi.iItem = i;
	lvi.pszText = buf;
	lvi.iSubItem = 0;

	if (win->layered) {
	    if (win->restored)
		sprintf(buf, "[%d%%]", (int)((((double)win->alpha) * 100.0) / 255.0));
	    else
		sprintf(buf, "%d%%", (int)((((double)win->alpha) * 100.0) / 255.0));
	} else
	    buf[0] = 0;

	SendMessage(listbox, LVM_SETITEMTEXT, i, (LPARAM)&lvi);
    }
}

void
update_window_list(HWND listbox)
{
    ListView_DeleteAllItems(listbox);

    // mark all items so we can check which ones were not refound and free them
    set<window *>::iterator it = g_all_windows.begin();
    while (it != g_all_windows.end())
	(*it++)->mark = 1;

    HWND w = GetWindow(GetDesktopWindow(), GW_CHILD);
    while (w) {
	char buf[1024];
	char buf2[1024];
	char buf3[1024];
	if (is_valid_window_to_show(w, buf, sizeof(buf),
				    buf2, sizeof(buf2))) {
	    window *win;
	    if (g_windows_by_hwnd.find(w) != g_windows_by_hwnd.end()) {
		win = g_windows_by_hwnd[w];
	    } else {
		string s = buf;
		s += "\001";
		s += buf2;
		if (g_windows_by_id.find(s) != g_windows_by_id.end()) {
		    win = g_windows_by_id[s];
		    win->set_hwnd(w);
		} else {
		    win = new window(w, buf, buf2);
		}
	    }
	    win->mark = 0;

	    LVITEM lvi;
	    int di;
	    ZeroMemory(&lvi, sizeof(lvi));

	    lvi.mask = LVIF_TEXT | LVIF_PARAM;
	    lvi.iItem = 0;
	    lvi.lParam = (LPARAM)win;

	    if (win->layered) {
		if (win->restored)
		    sprintf(buf3, "[%d%%]", (int)((((double)win->alpha) * 100.0) / 255.0));
		else
		    sprintf(buf3, "%d%%", (int)((((double)win->alpha) * 100.0) / 255.0));
	    } else
		buf3[0] = 0;

	    lvi.iSubItem = 0;
	    lvi.pszText = buf3;
	    di = ListView_InsertItem(listbox, &lvi);

	    lvi.iSubItem = 1;
	    lvi.pszText = buf;
	    SendMessage(listbox, LVM_SETITEMTEXT, di, (LPARAM)&lvi);

	    lvi.iSubItem = 2;
	    lvi.pszText = buf2;
	    SendMessage(listbox, LVM_SETITEMTEXT, di, (LPARAM)&lvi);
	}
	w = GetWindow(w, GW_HWNDNEXT);
    }

    it = g_all_windows.begin();
    while (it != g_all_windows.end()) {
	if ((*it)->mark == 1) {
	    // this one was not found, delete it!
	    g_windows_by_hwnd.erase((*it)->hwnd);
	    delete *it;
	    g_all_windows.erase(it++);
	} else
	    it++;
    }

    ListView_SortItems(listbox, listcmp, NULL);
}

bool g_minimize_on_start = 0;

int CALLBACK
dlgfunc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    HWND listbox = GetDlgItem(hwndDlg, IDC_WINLIST);

    switch (uMsg) {
	case WM_CLOSE:
	    {
		if (MessageBox(hwndDlg, "Closing Froost will cause it to "
			       "exit.\n"
			       "If you just wish to hide this window, please "
			       "minimize it.\n\n"
			       "Are you sure you want to quit this "
			       "application?",
			       "Closing Froost",
			       MB_YESNO) != IDYES)
		    return true;
		NOTIFYICONDATA tnid;
		tnid.cbSize = sizeof(NOTIFYICONDATA);
		tnid.hWnd = hwndDlg;
		tnid.uID = 1025;
		Shell_NotifyIcon(NIM_DELETE, &tnid);

		save();
	    }
	    break;

	case WM_USER + 1:
	    if (LOWORD(lParam) == WM_LBUTTONDOWN)
		ShowWindow(hwndDlg, SW_SHOWNORMAL);
	    break;

	case WM_SIZE:
	    if (wParam == SIZE_MINIMIZED)
		ShowWindow(hwndDlg, SW_HIDE);
	    break;

	case WM_INITDIALOG:
	    {
		NOTIFYICONDATA tnid;
		tnid.cbSize = sizeof(NOTIFYICONDATA);
		tnid.hWnd = hwndDlg;
		tnid.uID = 1025;
		tnid.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
		tnid.uCallbackMessage = WM_USER+1;
		tnid.hIcon = LoadIcon(GetModuleHandle(NULL),
				      MAKEINTRESOURCE(IDI_ICON));
		const char *tip = "Froost";
		lstrcpyn(tnid.szTip, tip, strlen(tip)+1);
		Shell_NotifyIcon(NIM_ADD, &tnid);

		// set icon
		SetClassLong(hwndDlg, GCL_HICON,
				  (long)LoadIcon(GetModuleHandle(NULL),
						 MAKEINTRESOURCE(IDI_ICON)));
		SetClassLong(hwndDlg, GCL_HICONSM,
				  (long)LoadIcon(GetModuleHandle(NULL),
						 MAKEINTRESOURCE(IDI_ICON)));

		SendDlgItemMessage(hwndDlg, IDC_TRANSPARENCY, TBM_SETRANGEMAX,
				   0, 255);
		SendDlgItemMessage(hwndDlg, IDC_TRANSPARENCY, TBM_SETRANGEMIN,
				   1, 0);
		SendDlgItemMessage(hwndDlg, IDC_TRANSPARENCY, TBM_SETTICFREQ,
				   16, 0);

		{
		    int i = LVS_EX_FULLROWSELECT;
#ifdef LVS_EX_LABELTIP
		    i |= LVS_EX_LABELTIP;
#endif
		    ListView_SetExtendedListViewStyle(listbox, i);
		}

		int n;
		LVCOLUMN lvc;
		lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT ;
		lvc.cx = 50;
		lvc.fmt = LVCFMT_LEFT;

		lvc.pszText = "Trans";
		n = ListView_InsertColumn(listbox, 0, &lvc);

		lvc.fmt = LVCFMT_LEFT;
		lvc.cx = 165;
		lvc.pszText = "Window Name";
		n = ListView_InsertColumn(listbox, 1, &lvc);

		lvc.fmt = LVCFMT_LEFT;
		lvc.cx = 150;
		lvc.pszText = "Window Class";
		n = ListView_InsertColumn(listbox, 2, &lvc);

		load();

		update_window_list(listbox);

	    }
	    SetTimer(hwndDlg, 0, 100, NULL);
	    return true;

	case WM_TIMER:
	    //XXX lame method to minimize on startup
	    if (g_minimize_on_start) {
		g_minimize_on_start = 0;
		ShowWindow(hwndDlg, SW_HIDE);
	    }
	    break;

	case WM_NOTIFY:
	    if (wParam == IDC_TRANSPARENCY) {
		NMHDR *nmhdr = (NMHDR*)lParam;
		if (nmhdr->code ==  NM_RELEASEDCAPTURE) {
		    int newpos = SendDlgItemMessage(hwndDlg, IDC_TRANSPARENCY,
						    TBM_GETPOS, 0, 0);

		    if (g_current) {
			if (!g_current->layered)
			    set_layered(g_current->hwnd, true);
			g_current->layered = 1;
			g_current->alpha = newpos;
			SetLayeredWindowAttributes(g_current->hwnd, 0,
						   g_current->alpha, LWA_ALPHA);
			update_window_item(listbox, g_current);
		    }
		}

	    } else if (wParam == IDC_WINLIST) {
		NMHDR *nmhdr = (NMHDR*)lParam;
		if (nmhdr->code ==  LVN_ITEMCHANGED) {
		    NMLISTVIEW *nmlv = (NMLISTVIEW *)nmhdr;
		    if (nmlv->uNewState & LVIS_SELECTED) {
			LVITEM lvi;
			ZeroMemory(&lvi, sizeof(lvi));
			lvi.iItem = nmlv->iItem;
			lvi.mask = LVIF_PARAM;
			ListView_GetItem(listbox, &lvi);

			window *w = (window*)lvi.lParam;

			g_current = w;

			string s;
			s += "\"";
			s += w->title;
			s += "\"";
			s += " (";
			s += w->classname;
			s += ")";
			SetDlgItemText(hwndDlg, IDC_WINNAME, s.c_str());

			if (w->restored)
			    CheckDlgButton(hwndDlg, IDC_RESTORE, BST_CHECKED);
			else
			    CheckDlgButton(hwndDlg, IDC_RESTORE, BST_UNCHECKED);

			if (w->layered) {
			    CheckDlgButton(hwndDlg, IDC_ENABLE, BST_CHECKED);
			    SendDlgItemMessage(hwndDlg, IDC_TRANSPARENCY,
					       TBM_SETPOS, 1, w->alpha);
			    EnableWindow(GetDlgItem(hwndDlg, IDC_TRANSPARENCY), 1);
			    EnableWindow(GetDlgItem(hwndDlg, IDC_RESTORE), 1);
			} else {
			    CheckDlgButton(hwndDlg, IDC_ENABLE, BST_UNCHECKED);
			    SendDlgItemMessage(hwndDlg, IDC_TRANSPARENCY,
					       TBM_SETPOS, 1, 255);
			    EnableWindow(GetDlgItem(hwndDlg, IDC_TRANSPARENCY), 0);
			    EnableWindow(GetDlgItem(hwndDlg, IDC_RESTORE), 0);
			}
			update_window_item(listbox, w);
		    }
		}
	    }
	    break;

	case WM_COMMAND:
	    switch(HIWORD(wParam)) {
		case BN_CLICKED:
		    {
			switch (LOWORD(wParam)) {
			    case IDCANCEL:
				EndDialog(hwndDlg, wParam);
				return true;

			    case IDC_UPDATE:
				update_window_list(GetDlgItem(hwndDlg, IDC_WINLIST));
				return true;

			    case IDC_RESTORE:
				if (!g_current) break;
				if (IsDlgButtonChecked(hwndDlg, IDC_RESTORE))
				    g_current->restored = 1;
				else
				    g_current->restored = 0;
				update_window_item(listbox, g_current);
				return true;

			    case IDC_ENABLE:
				if (!g_current) break;
				if (IsDlgButtonChecked(hwndDlg, IDC_ENABLE)) {
				    if (!g_current->layered) {
					set_layered(g_current->hwnd, true);
					g_current->layered = 1;
					g_current->alpha = 255;
					g_current->restored = 0;
					SetLayeredWindowAttributes(g_current->hwnd, 0,
								   g_current->alpha, LWA_ALPHA);
					SendDlgItemMessage(hwndDlg, IDC_TRANSPARENCY,
							   TBM_SETPOS, 1, g_current->alpha);
					EnableWindow(GetDlgItem(hwndDlg, IDC_TRANSPARENCY), 1);
					EnableWindow(GetDlgItem(hwndDlg, IDC_RESTORE), 1);
				    }
				} else {
				    if (g_current->layered) {
					set_layered(g_current->hwnd, false);
					g_current->layered = 0;
					g_current->restored = 0;
					SendDlgItemMessage(hwndDlg, IDC_TRANSPARENCY,
							   TBM_SETPOS, 1, 255);
					EnableWindow(GetDlgItem(hwndDlg, IDC_TRANSPARENCY), 0);
					EnableWindow(GetDlgItem(hwndDlg, IDC_RESTORE), 0);
				    }
				}
				update_window_item(listbox, g_current);
				return true;
			}
		    }
	    }
    }
    return false;
}

int APIENTRY
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	LPSTR lpCmdLine, int nCmdShow)
{
    if (!strcmp(lpCmdLine, "-minimize"))
	g_minimize_on_start = 1;

    INITCOMMONCONTROLSEX ic;
    ic.dwSize = sizeof(ic);
    ic.dwICC = ICC_TREEVIEW_CLASSES | ICC_LISTVIEW_CLASSES;
    InitCommonControlsEx(&ic);

    if (RegCreateKeyEx(HKEY_CURRENT_USER,
		       "Software\\Ishiboo\\Froost", NULL, NULL, 0,
		       KEY_ALL_ACCESS, NULL, &regkey, NULL) != ERROR_SUCCESS) {
       MessageBox(NULL, NULL, "FATAL: Could not create registry key",
                  MB_ICONWARNING | MB_OK);
       return 0;
    }

    WNDCLASSEX wc;
    ZeroMemory(&wc, sizeof(wc));
    if (!GetClassInfoEx(GetModuleHandle(NULL), "froost_container", &wc)) {
	wc.cbSize        = sizeof(wc);
	wc.style         = 0;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = GetModuleHandle(NULL);
	wc.hIcon         = LoadIcon(NULL, MAKEINTRESOURCE(IDI_ICON));
	wc.hCursor       = NULL;
	wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = "froost_container";
	wc.lpfnWndProc   = DefWindowProc;
	ATOM a = RegisterClassEx(&wc);
	assert(a);
    }
    HWND prnt = CreateWindow("froost_container", "",
			     0, 0,0,0,0, NULL, NULL,
			     GetModuleHandle(NULL), NULL);

    if (DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAINDLG), prnt, dlgfunc) < 0) {
	char buf[1000];
	sprintf(buf, "error creating dialog: %d\n", GetLastError());
	MessageBox(NULL, buf, "error", MB_ICONWARNING | MB_OK);
    }
    return 0;
}

window::window(HWND w, const char *tit, const char *clsnm)
{
    g_all_windows.insert(this);
    layered = is_layered(w, &alpha);
    set_hwnd(w);
    restored = 0;
    title = tit;
    classname = clsnm;
    mark = 0;

    string s = tit;
    s += "\001";
    s += clsnm;
    g_windows_by_id[s] = this;
}

window::window(const char *tit, const char *clsnm, BYTE alph)
{
    g_all_windows.insert(this);
    hwnd = 0;
    layered = 1;
    alpha = alph;
    restored = 1;
    title = tit;
    classname = clsnm;
    mark = 0;

    string s = tit;
    s += "\001";
    s += clsnm;
    g_windows_by_id[s] = this;
}

void
window::set_hwnd(HWND w)
{
    g_windows_by_hwnd[w] = this;
    hwnd = w;
    if (layered) {
	set_layered(hwnd, 1);
	SetLayeredWindowAttributes(hwnd, 0, alpha, LWA_ALPHA);
    }
}
